
#include <unistd.h>
#include <ctype.h>

#include "globdef.h"
#include "uidef.h"
#include "fft1def.h"
#include "fft2def.h"
#include "fft3def.h"
#include "sigdef.h"
#include "screendef.h"
#include "seldef.h"
#include "thrdef.h"
#include "txdef.h"
#include "vernr.h"


#define TX_WRBLOCK_MAXCNT 50
#define TX_AGC_THRESHOLD 1.

#define TX_SETUP_AD 0
#define TX_SETUP_MUTE 1
#define TX_SETUP_MICAGC 2
#define TX_SETUP_RF1AGC 3
#define TX_SETUP_RF1CLIP 4

#define TX_BAR_AD_LINE 8
#define TX_BAR_MUTE_LINE 10 
#define TX_BAR_MICAGC_LINE 12
#define TX_BAR_RF1AGC_LINE 14
#define TX_BAR_RF1CLIP_LINE 16
#define TX_BAR_ALC_LINE 18

#define TXPAR_INPUT_LINE 21

#define TX_BAR_SIZE 256

#define MAX_DISPLAY_TYPES 4
#define DISPLAY_HOLD_SIZE 8

int tx_pixels_per_bin;
short int *txtrace;
float txad_hz_per_bin;
float txtrace_gain;
int txtrace_height;
float tx_fft1bin_minpower;
float tx_fft1_minpower;
float *mic_filter;
float *tx_fft_win1;
float *tx_fft_win2;
float *tx_rf2_pwrf;
float *tx_rf2_pwrd;
float *tx_iftmp1;
float *tx_iftmp2;
int *tx_out_int;
short int *tx_out_shi;

float tx_agc_factor;
float tx_agc_decay;
float tx_fft_block_time;
float txpwr_decay;
float tx_forwardpwr;
float tx_backpwr;
int tx_rf1_pa;
int tx_rf1_px;
int tx_fft2_pa;
int tx_fft2_px;
int tx_rf2_pa;
int tx_rf2_pb;
int tx_sizhalf1;
int tx_sizhalf2;
int tx_sizhalf3;
int tx_output_flag;
float tx_resample_ratio;
float new_tx_resample_ratio;
int tx_output_upsample;
float tx_rf2_fx;
int tx_if1_pa;
int tx_if1_px;
int tx_ifbuf_size;
float tx_if_phase;
float tx_if_phstep;
char display_color[MAX_DISPLAY_TYPES]={0,10,11,12};
int tx_pre_upsample;
int tx_final_upsample;
int tx_fftsize3;
int tx_wrblock_no;
int tx_outbuf_ref;
int tx_outbuf_current;

void show_txfft(float *z, float lim, int type);

void check_txparms(void)
{
int i,k;
float t1;
if(tx.lowest_bin < 1)tx.lowest_bin=1;
if(tx.lowest_bin > 3000/txad_hz_per_bin)tx.lowest_bin=3000/txad_hz_per_bin;
if(tx.highest_bin < 1000/txad_hz_per_bin)tx.highest_bin=1000/txad_hz_per_bin;
if(tx.highest_bin > tx_fftsize1/2-2)tx.highest_bin=tx_fftsize1/2-2;
if(tx.highest_bin > 8000/txad_hz_per_bin)tx.highest_bin=8000/txad_hz_per_bin;
if(tx.highest_bin < tx.lowest_bin+1)tx.highest_bin=tx.lowest_bin+1;
t1=(tx.highest_bin-tx.lowest_bin+1)*txad_hz_per_bin;
if(tx.output_channels == 1)t1*=2;
if(t1 > 0.75*tx.output_speed)
  {
  t1=tx.output_speed;
  if(tx.output_channels == 1)t1/=2;
  t1/=txad_hz_per_bin;
  tx.highest_bin=tx.lowest_bin+t1;
  }
k=tx_fftsize1/tx.highest_bin;
tx_fftn2=tx_fftn1;
tx_fftsize2=tx_fftsize1;
tx_rfbuf_size2=tx_rfbuf_size1;
tx_pre_upsample=1;
while(k<8)
  {
  tx_pre_upsample*=2;
  k*=2;
  tx_fftn2++;
  tx_fftsize2*=2;
  tx_rfbuf_size2*=2;
  }
tx_rfbuf_mask2=tx_rfbuf_size2-1;
tx_sizhalf2=tx_fftsize2/2;      
tx_fft_block2=2*tx_fftsize2;
init_fft(0,tx_fftn2, tx_fftsize2, tx_fft_table2, tx_fft_permute2);
make_window(2,tx_sizhalf2, 2, tx_fft_win2);
t1=1/(1.62*tx_fftsize2);
for(i=0; i<tx_sizhalf2; i++)
  {
  tx_fft_win2[i]*=t1;
  }
tx_final_upsample=tx_output_upsample/tx_pre_upsample;
for(i=0; i<tx_fftsize1; i++)mic_filter[i]=0;
if(tx.slope > 50 || tx.slope < -50)tx.slope=0;
if(tx.bass > 50 || tx.bass < -50)tx.bass=0;
if(tx.treble > 50 || tx.treble < -50)tx.treble=0;
// Store the desired frequency response in dB.
t1=0;
for(i=tx.lowest_bin; i<=tx.highest_bin; i++)
  {
  mic_filter[i]=t1;
  t1+=0.001*tx.slope*txad_hz_per_bin;
  }
k=(tx.highest_bin-tx.lowest_bin)/4;
t1=0;
for(i=tx.lowest_bin+k; i>=tx.lowest_bin; i--)
  {
  t1+=0.001*tx.bass*txad_hz_per_bin;
  mic_filter[i]+=t1;
  }
t1=0;
for(i=tx.highest_bin-k; i<=tx.highest_bin; i++)
  {
  t1+=0.001*tx.treble*txad_hz_per_bin;
  mic_filter[i]+=t1;
  }
// Convert from dB to linear scale (amplitude) and normalize
// the filter curve.
t1=0;
for(i=tx.lowest_bin; i<=tx.highest_bin; i++)
  {
  mic_filter[i]=pow(10.,0.05*mic_filter[i]);
  t1+=mic_filter[i]*mic_filter[i];
  }
t1=1/sqrt(t1/(tx.highest_bin-tx.lowest_bin+1));
for(i=tx.lowest_bin; i<=tx.highest_bin; i++)
  {
  mic_filter[i]*=t1;
  }
if(tx.mic_f_threshold < 0)tx.mic_f_threshold=0;
if(tx.mic_f_threshold > 120)tx.mic_f_threshold=120;
if(tx.mic_t_threshold < 0)tx.mic_t_threshold=0;
if(tx.mic_t_threshold > 120)tx.mic_t_threshold=120;
tx_fft1bin_minpower=pow(10.,0.1*(tx.mic_f_threshold-80));
tx_fft1_minpower=pow(10.,0.1*(tx.mic_t_threshold+tx.mic_f_threshold-80));
tx_agc_factor=1;
tx_agc_decay=pow(2.718,-100.0*tx_fft_block_time/tx.mic_agc_time);
txpwr_decay=pow(2.718,-1/(tx.input_speed*0.025));
tx_forwardpwr=0;
tx_backpwr=0;
}


void use_tx_if(void)
{
char s[80];
int i, k, m;
float t1, t2;
// The latest half block of data resides in tx_if1[tx_fftsize2]
// to tx_if1[2*tx_fftsize2-1]
// Copy the previous block from tx_iftmp1 and multiply with
// the window function.
k=tx_sizhalf2;
m=k-1;
for(i=0; i<tx_sizhalf2; i++)
  {
  tx_if1[2*i  ]=tx_iftmp1[2*i  ]*tx_fft_win2[i]; 
  tx_if1[2*i+1]=tx_iftmp1[2*i+1]*tx_fft_win2[i]; 
  tx_iftmp1[2*i  ]=tx_if1[2*k  ];
  tx_iftmp1[2*i+1]=tx_if1[2*k+1];
  tx_if1[2*k  ]*=tx_fft_win2[m];
  tx_if1[2*k+1]*=tx_fft_win2[m];
  m--;
  k++;
  } 
fftforward(tx_fftsize2, tx_fftn2, tx_if1, tx_fft_table2, tx_fft_permute2);
// Clear the spectrum outside the desired passband.
// These data points should be very small. 
k=tx.lowest_bin/tx_resample_ratio-2;
if(k<1)k=1;
m=2+(tx.highest_bin+1)/tx_resample_ratio;
//fprintf(dmp,"\ntx_resample_ratio %f, tx_output_upsample %d ",
//                                   tx_resample_ratio,tx_output_upsample);
for(i=0; i<k; i++)
  {
  tx_if1[2*i]=0;
  tx_if1[2*i+1]=0;
  }
for(i=m; i<tx_fftsize3; i++)
  {
  tx_if1[2*i]=0;
  tx_if1[2*i+1]=0;
  }
show_txfft(tx_if1,0,2);
// Now we have the output in the frequency domain at the size tx_fftsize3
// with a sampling rate that fits out D/A converter. Transform it back
// to the time domain into tx_if2.
fftback(tx_fftsize3, tx_fftn3, tx_if1, tx_fft_table3, tx_fft_permute3);
lir_sched_yield();
for(i=0; i<tx_fftsize3; i+=2)
  {
  tx_if2[i  ]=tx_if1[i  ]+tx_iftmp2[i  ];
  tx_if2[i+1]=tx_if1[i+1]+tx_iftmp2[i+1];
  }
tx_if2[0]=2;//ö
for(i=0; i<tx_fftsize3; i++)
  {
  tx_iftmp2[i]=tx_if1[tx_fftsize3+i];
  }
// Shift the frequency and convert the data to fit the output format.
lir_sched_yield();
if(tx.output_bytes==4)
  {
  if(tx.output_channels==1)
    {
    k=4;
    for(i=0; i<tx_sizhalf3; i++)
      {
      t1=tx_if2[2*i  ];
      t2=tx_if2[2*i+1];
      tx_out_int[i]=0x7f000000*t1;
      }
    write(tx_audio_out,tx_out_int,sizeof(int)*tx_sizhalf3);
    }
  else
    {
    k=8;  
    for(i=0; i<tx_sizhalf3; i++)
      {
      t1=tx_if2[2*i  ];
      t2=tx_if2[2*i+1];
      tx_out_int[2*i  ]=0x3f000000*t1;
      tx_out_int[2*i+1]=0x3f000000*t2;
      }
    write(tx_audio_out,tx_out_int,sizeof(int)*tx_fftsize3);
    }      
  }
else
  {  
  if(tx.output_channels==1)
    {
    k=2;
    for(i=0; i<tx_sizhalf3; i++)
      {
      t1=tx_if2[2*i  ];
      t2=tx_if2[2*i+1];
      tx_out_shi[i]=0x7f00*t1;
      }
    write(tx_audio_out,tx_out_shi,tx_sizhalf3);
    }
  else
    {  
    k=4;
    for(i=0; i<tx_sizhalf3; i++)
      {
      t1=tx_if2[2*i  ];
      t2=tx_if2[2*i+1];
      tx_out_shi[2*i]=0x7f00*t1;
      tx_out_shi[2*i+1]=0x7f00*t2;
      }

    write(tx_audio_out,tx_out_shi,tx_fftsize3);


    }
  }
tx_wrblock_no++;
tx_outbuf_current+=lir_tx_output_bytes();
if(tx_wrblock_no == TX_WRBLOCK_MAXCNT)
  {
  tx_wrblock_no=0;
  if(tx_outbuf_ref == 0)
    {
    tx_outbuf_ref=tx_outbuf_current;
    }
  else
    {
    t1=tx_outbuf_current-tx_outbuf_ref;
    tx_outbuf_current=0;
// We have accumulated an error of t1/TX_WRBLOCK_MAXCNT bytes while writing 
// k*TX_WRBLOCK_MAXCNT bytes. Therefore we should multiply the output 
// speed by 1+t1/(k*TX_WRBLOCK_MAXCNT*TX_WRBLOCK_MAXCNT*sizhalf3.
    new_tx_resample_ratio=tx_resample_ratio*(1+t1/(2*tx_pre_upsample*
       tx_final_upsample*k*TX_WRBLOCK_MAXCNT*TX_WRBLOCK_MAXCNT*tx_sizhalf3));
    sprintf(s,"%f  %f  ",t1,new_tx_resample_ratio);
lir_text(0,0,s);
    }
  }
lir_refresh_screen();
tx_if1_pa=0;
}


void resample_tx_output(void)
{
int i1, i2, i3, i4, mask2;
float t1, t2, t3, t4, t5, t6, t7, r1, r2;
float a1, a2, rdiff, resamp;
// The microphone is best sampled at a low sampling speed since
// we do not want much bandwidth nor dynamic range.
// For the output the requirements depend on the circumstances.
// In case the transmitter is used in a wideband up-conversion
// system, the output sampling rate may be very high and the
// requirements on the signal purity might be very high.
// We do the resampling in two steps. First do a fractional
// resampling to a speed near the speed of the microphone
// signal (but perhaps a power of two below the desired output
// speed as indicated by tx_output_upsample.)
// Subsequently, use an FFT to sample up the signal to the
// desired frequency while filtering out the false signals
// that non-linearities in the fractional resampling have
// introduced.
resamp=tx_resample_ratio*tx_pre_upsample;
mask2=tx_rfbuf_mask2-1;
begin:;
r1=tx_if1_pa/tx_resample_ratio;
resamp:;
r2=(tx_if1_pa+2)/tx_resample_ratio;
i2=tx_rf2_fx+r1+1;
i3=tx_rf2_fx+r2+1;
i2&=mask2;
i3&=mask2;
if(abs(i3-i2) > 2)
  {
  i2=tx_rf2_fx+(r1+r2)/2;
  i3=i2+2;
  i2&=mask2;
  i3&=mask2;
  }
else
  {
  if(i3==i2)
    {
    i2=tx_rf2_fx+r1;
    i2&=mask2;
    if(i3==i2)
      {
      i3=i2+2;
      i3&=mask2;
      }
    }
  }        
i4=(i3+2)&(mask2);
i1=(i2+mask2)&(mask2);
if( ((tx_rf2_pb-i4+tx_rfbuf_size2)&tx_rfbuf_mask2) > 4)
  {
  rdiff=r1+tx_rf2_fx-i2;
  if(rdiff > tx_rfbuf_size2/2)
    {
    rdiff-=tx_rfbuf_size2;
    }
  rdiff/=2; 
  tx_if_phase+=tx_if_phstep;
  if(tx_if_phase >  PI_L)tx_if_phase-=2*PI_L;
  if(tx_if_phase < -PI_L)tx_if_phase+=2*PI_L;
// Use Lagrange's interpolation formula to fit a third degree
// polynomial to 4 points:
//  a1=-rdiff *   (rdiff-1)*(rdiff-2)*tx_rf2[i1]/6
//     +(rdiff+1)*(rdiff-1)*(rdiff-2)*tx_rf2[i2]/2
//     -(rdiff+1)* rdiff   *(rdiff-2)*tx_rf2[i3]/2
//     +(rdiff+1)* rdiff   *(rdiff-1)*tx_rf2[i4]/6; 
// Rewrite slightly to save a few multiplications - do not
// think the compiler is smart enough to do it for us.
  t1=rdiff-1;
  t2=rdiff-2;
  t3=rdiff+1;
  t4=t1*t2;
  t5=t3*rdiff;
  t6=rdiff*t4;
  t4=t3*t4;
  t7=t5*t2;
  t5=t5*t1;
  a1=((t5*tx_rf2[i4  ]-t6*tx_rf2[i1  ])/3+t4*tx_rf2[i2  ]-t7*tx_rf2[i3  ])/2;
  a2=((t5*tx_rf2[i4+1]-t6*tx_rf2[i1+1])/3+t4*tx_rf2[i2+1]-t7*tx_rf2[i3+1])/2;
  tx_if1[tx_if1_pa+tx_fftsize2  ]=a1;
  tx_if1[tx_if1_pa+tx_fftsize2+1]=a2;
  r1=r2;
  tx_if1_pa+=2;
  if(tx_if1_pa >= tx_fftsize2)
    {
    tx_rf2_fx+=r2;
    tx_resample_ratio=new_tx_resample_ratio;
    if(tx_rf2_fx>tx_rfbuf_size2)tx_rf2_fx-=tx_rfbuf_size2;
    use_tx_if();
    goto begin;
    }
  goto resamp;  
  }
}



void tx_bar(int xt,int yt,int val1, int val2)
{
int x, y;
x=xt*text_width;
y=yt*text_height+1;
if(val1 == val2)
  {
  if(val1 >= 0)return;
  lir_fillbox(x,y,TX_BAR_SIZE,text_height-2,7);
  return;
  }
if(val2 > val1)
  { 
  lir_fillbox(x+val1,y,val2-val1,text_height-2,12);
  }
else
  {  
  lir_fillbox(x+val2,y,val1-val2,text_height-2,7);
  } 
}


void show_txfft(float *z, float lim, int type)
{
int i, ia, ib, k;
float t1,t2;
char color;
short int *trc;
if(type >= MAX_DISPLAY_TYPES)
  {
  lirerr(762319); 
  return;
  }
trc=&txtrace[type*tx_fftsize1];  
ia=0;
t2=0;
ib=tx_pixels_per_bin;
lir_sched_yield();
for(i=1; i<tx_fftsize1; i++)
  {
if(type != 0)
  lir_line(ia,trc[i-1],ib,trc[i],0);
  ia=ib;
  ib+=tx_pixels_per_bin;
  }
lir_sched_yield();
ia=0;
ib=tx_pixels_per_bin;

for(i=0; i<tx_fftsize1; i++)
  {
  t1=z[2*i]*z[2*i]+z[2*i+1]*z[2*i+1];
  k=-100*log10(txtrace_gain*(t1+0.0000000000001));
  if(k>txtrace_height)k=txtrace_height;
  if(k<0)k=0;
  trc[i]=k+screen_height-txtrace_height-1;
  if(i>0)
    {
    if(type==0)
      {
      if(t1 > lim || t2 > lim)
        {
        color=15;
        }
      else
        {
        color=12;
        }
      }
    else
      {
      color=display_color[type];
      }      
    lir_line(ia,trc[i-1],ib,trc[i],color);
    ia=ib;
    ib+=tx_pixels_per_bin;
    }
  t2=t1;
  }
lir_sched_yield();
}



void spproc_setup(void)
{
char s[80];
int i, j, k, pb;
int setup_mode;
int ad_pix1;
int ad_pix2;
int mute_pix1;
int mute_pix2;
int micagc_pix1;
int micagc_pix2;
int rf1agc_pix1;
int rf1agc_pix2;
int rf1clip_pix1;
int rf1clip_pix2;
int alc_pix1;
int alc_pix2;
float t1, t2, r1, r2;
float prat, pmax;
int old_display_ptr;
int rf2ampl_pa;
int old_admax[DISPLAY_HOLD_SIZE];
float old_rf1agc[DISPLAY_HOLD_SIZE];
float old_rf1clip[DISPLAY_HOLD_SIZE];
float old_prat1[DISPLAY_HOLD_SIZE];
float old_prat2[DISPLAY_HOLD_SIZE];
open_tx_output();
init_txmem();
if(kill_all_flag)return;
open_tx_input();
if(kill_all_flag)return;
tx_pixels_per_bin=(screen_width-1)/tx_fftsize1;
if(tx_pixels_per_bin == 0)
  {
  lirerr(778543);
  return;
  }
txtrace=malloc(MAX_DISPLAY_TYPES*tx_fftsize1*sizeof(short int));
if(txtrace == NULL)
  {
  lirerr(778543);
  return;
  }
for(i=0; i<MAX_DISPLAY_TYPES*tx_fftsize1; i++)txtrace[i]=screen_height-1;
txtrace_gain=0.2;
txtrace_height=screen_height-text_height*(TXPAR_INPUT_LINE+1);
lir_sleep(50000);
linrad_thread_create(THREAD_TX_INPUT);
setup_mode=TX_SETUP_AD;
restart:;
tx_fft1_px=tx_fft1_pa;
tx_rf1_pa=0;
tx_rf1_px=0;
tx_fft2_pa=0;
tx_fft2_px=0;
tx_rf2_pa=0;
tx_rf2_pb=0;
tx_rf2_fx=0;
old_display_ptr=0;
for(i=0; i<DISPLAY_HOLD_SIZE; i++)
  {
  old_admax[i]=0;
  old_rf1agc[i]=0;
  old_rf1clip[i]=0;
  old_prat1[i]=90;
  old_prat2[i]=90;
  }
clear_screen();
tx_ad_maxamp=0;
ad_pix1=0;
ad_pix2=0;
micagc_pix1=0;
micagc_pix2=0;
rf1agc_pix1=0;
rf1agc_pix2=0;
mute_pix1=0;
mute_pix2=0;
rf1clip_pix1=0;
rf1clip_pix2=0;
alc_pix1=0;
alc_pix2=0;
settextcolor(14);
lir_text(0,TX_BAR_AD_LINE-2,"Press 'S' to save parameters.");
lir_text(38,TX_BAR_AD_LINE-2,"'+' or '-' to change spectrum scale.");
lir_text(0,TX_BAR_AD_LINE,"'M' =");
lir_text(0,TX_BAR_MUTE_LINE,"'Q' =");
lir_text(0,TX_BAR_MICAGC_LINE,"'A' =");
lir_text(0,TX_BAR_RF1CLIP_LINE,"'C' =");
settextcolor(7);
lir_text(6,TX_BAR_AD_LINE,"Soundcard");
lir_text(6,TX_BAR_MUTE_LINE,"Mute");
tx_bar(16,TX_BAR_AD_LINE,-1,-1);
tx_bar(16,TX_BAR_MUTE_LINE,-1,-1);
lir_text(6,TX_BAR_MICAGC_LINE,"MIC AGC");
tx_bar(16,TX_BAR_MICAGC_LINE,-1,-1);
lir_text(6,TX_BAR_RF1AGC_LINE,"RF1 AGC");
tx_bar(16,TX_BAR_RF1AGC_LINE,-1,-1);
lir_text(6,TX_BAR_RF1CLIP_LINE,"RF clip");
tx_bar(16,TX_BAR_RF1CLIP_LINE,-1,-1);
lir_text(6,TX_BAR_ALC_LINE,"ALC");
tx_bar(16,TX_BAR_ALC_LINE,-1,-1);
switch (setup_mode)
  {
  case TX_SETUP_AD:
  lir_text(0,0,
            "Verify AF clip level, (some device drivers give too few bits)");
  lir_text(0,1,"Use mixer program to set volume for the AF to never reach");
  lir_text(0,2,"its maximum during normal operation.");
  lir_text(0,3,
              "Press 'L' or 'H' to change cut-off frequencies,'F','B' or 'T'");
  lir_text(0,4,"for slope, bass or treble to change frequency response.");
  sprintf(s,
    "Low %.0f Hz  High %.0f Hz   Slope %d dB/kHz   Bass %d dB  Treble %d dB",
       tx.lowest_bin*txad_hz_per_bin, tx.highest_bin*txad_hz_per_bin,
                                            tx.slope, tx.bass, tx.treble);
  lir_text(0,5,s);
  break;

  case TX_SETUP_MUTE:
  lir_text(0,0,"Press 'F' to set frequency domain threshold or");
  lir_text(0,1,"Press 'T' to set time domain threshold for signal");
  
  lir_text(0,2,"to become muted when only background noise is present.");
  sprintf(s,"Thresholds:  F=%d dB   T=%d dB",
                                 tx.mic_f_threshold, tx.mic_t_threshold);
  lir_text(0,3,s);
  break;
  
  case TX_SETUP_MICAGC:
  lir_text(0,0,"Press 'V' to set mic volume for suitable AGC action");
  lir_text(0,1,"or 'T' to set decay time constant");
  sprintf(s,"Vol=%d dB   T=%.2f sek",tx.mic_gain, 0.01*tx.mic_agc_time);
  lir_text(0,2,s);
  break;
  
  case TX_SETUP_RF1CLIP:
  lir_text(0,0,"Press 'B' to set RF1 gain for desired clip level.");
  lir_text(0,1,"Bar range is 0 to 30 dB");
  sprintf(s,"RF1 gain=%d dB",tx.rf1_gain);
  lir_text(0,2,s);
  break;
  }
  
while(!kill_all_flag)
  {
  old_display_ptr++;
  if(old_display_ptr>=DISPLAY_HOLD_SIZE)old_display_ptr=0;
  memcheck(2,txmem,(int*)&txmem_handle);
// Show the microphone input level. 
  ad_pix2=tx_ad_maxamp;
  if(tx.input_bytes != 2)
    {
    ad_pix2/=0x10000;
    }  
  old_admax[old_display_ptr]=ad_pix2;
  t1=0;
  for(i=0; i<DISPLAY_HOLD_SIZE; i++)
    {
    if(t1<old_admax[i])t1=old_admax[i];
    }
  t1=20*log10(t1/0x8000);  
  sprintf(s,"%.1fdB  ",t1);
  lir_text(18+TX_BAR_SIZE/text_width, TX_BAR_AD_LINE, s);
  ad_pix2=(ad_pix2*TX_BAR_SIZE)/0x8000;    
  tx_bar(16,TX_BAR_AD_LINE,ad_pix1,ad_pix2);
  ad_pix1=ad_pix2;
  tx_ad_maxamp=0.95*tx_ad_maxamp;
// **************************************************************
// The setup routine contains the same processing steps as the
// transmit routine, but it has varoius display functions added
// into it. The processing steps are numbered and described below.
// **************************************************************
//
// ************* Step 1. *************
// Wait for a new block of data from the mic input routine.
// Note that data arrives as fourier transforms (tx_fft1) and
// that the mix input routine tx_input() already has applied
// the filtering of the microphone signal.
  lir_sem_wait(SEM_TXINPUT);
  show_txfft(&tx_fft1[tx_fft1_px],tx_fft1bin_minpower,0);
// ************* Step 2. *************
// Compute the power for each bin within the passband.
// Apply muting in the frequency domain by setting bins with too small
// signal power to zero.
// Collect the sum of the powers from surviving bins and mute the entire
// block in case the total power is below threshold.
// Here we use the agc factor as determined from the the
// previous block and attenuated with the decay factor.
// Note that the agc factor is limited to 20 dB at this point.
// A big pulse will not kill the signal for a long time!!
  tx_agc_factor=tx_agc_decay*tx_agc_factor+(1-tx_agc_decay);
  r1=1/(tx_agc_factor*tx_agc_factor);
  t1=0;
  k=0;
  for(i=tx.lowest_bin; i<=tx.highest_bin; i++)
    {
    t2=tx_fft1[tx_fft1_px+2*i  ]*tx_fft1[tx_fft1_px+2*i  ]+
       tx_fft1[tx_fft1_px+2*i+1]*tx_fft1[tx_fft1_px+2*i+1];
    if(t2*r1 > tx_fft1bin_minpower)
      {
      k++;
      t1+=t2;
      }
    else
      {
      tx_fft1[tx_fft1_px+2*i  ]=0;
      tx_fft1[tx_fft1_px+2*i+1]=0;
      }
    }
  if(k>0 && t1*r1 < tx_fft1_minpower)
    {
    for(i=tx.lowest_bin; i<=tx.highest_bin; i++)
      {
      tx_fft1[tx_fft1_px+2*i  ]=0;
      tx_fft1[tx_fft1_px+2*i+1]=0;
      }
    k=0;
    }
  if(k==0)
    {
    sprintf(s,"MUTED");
    }
  else
    {
    sprintf(s,"     ");
// ************* Step 3. *************
// Find out what the average amplitude would become if the old AGC
// factor would be applied. Change the agc factor in case the
// average amplitude is above the limit.
// Note that this step is only executed if k!=0.
// When k==0, the entire block is muted and the operation
// would be meaningless on a zero transform.
    t1=sqrt(t1);
    r1=pow(10.0,0.05*tx.mic_gain);
    t2=r1*t1*tx_agc_factor*(tx.highest_bin-tx.lowest_bin+1);
    if(t2 > TX_AGC_THRESHOLD*(tx.highest_bin-tx.lowest_bin+1))
      {
      tx_agc_factor=TX_AGC_THRESHOLD/(r1*t1);
      }
   t2=tx_agc_factor*(tx.highest_bin-tx.lowest_bin+1);
// Apply the AGC within the current fft block.
    for(i=tx.lowest_bin; i<=tx.highest_bin; i++)
      {
      tx_fft1[tx_fft1_px+2*i  ]*=t2;
      tx_fft1[tx_fft1_px+2*i+1]*=t2;
      }
    }
  mute_pix2=(k*TX_BAR_SIZE)/(tx.highest_bin-tx.lowest_bin+1);    
  tx_bar(16,TX_BAR_MUTE_LINE,mute_pix1,mute_pix2);
  mute_pix1=mute_pix2;
  settextcolor(15);  
  lir_text(18+TX_BAR_SIZE/text_width,TX_BAR_MUTE_LINE,s);
  settextcolor(7);      
// In case AGC has reduced the gain by more than 20 dB, set the
// AGC factor to -20 dB immediately because voice should never
// be as loud. This is to avoid the agc release time constant
// for impulse sounds caused by hitting the microphone etc.       
  t1=20*log10(tx_agc_factor);  
  sprintf(s,"%.1fdB  ",t1);
  lir_text(18+TX_BAR_SIZE/text_width, TX_BAR_MICAGC_LINE, s);
  if(tx_agc_factor < 0.1)tx_agc_factor=0.1;
  micagc_pix2=-log10(tx_agc_factor)*TX_BAR_SIZE;
  tx_bar(16,TX_BAR_MICAGC_LINE,micagc_pix1,micagc_pix2);
  micagc_pix1=micagc_pix2;
// ************* Step 4. *************
// Go back to the time domain and store the signal in tx_rf1
// Remember that we use sin squared windows and that
// the transforms overlap with 50%.
//
// Ideally, the peak amplitude should be 1.0, the audio AGC
// should be active and keep the power level nearly constant.
// For a voice signal the waveform will differ from time to time
// and therefore the peak to average power ratio will vary.
// Compute the peak amplitude for the current transform
// and save it for the next time we arrive here.
// Use the peak amplitude for an AGC in the time domain (Hilbert space).
// The Hilbert space AGC is a constant that may vary from one
// FFT block to the next one. It is equivalent with an AM modulator
// with a bandwidth corresponding to a single bin in the FFT so this
// AM modulation will not increase the bandwidth notably, but it
// will bring the RF amplitude near unity always.
//
// Finally, apply an amplitude limiter to the complex time domain signal.
// It will work as an RF limiter on the SSB signal and cause a lot of
// signal outside the passband.
  fftback(tx_fftsize1, tx_fftn1, &tx_fft1[tx_fft1_px], 
                                              tx_fft_table1, tx_fft_permute1);
  lir_sched_yield();
  r2=0;
  for(i=0; i<tx_fftsize1; i++)
    {
    t2=tx_fft1[tx_fft1_px+2*i  ]*tx_fft1[tx_fft1_px+2*i  ]+
       tx_fft1[tx_fft1_px+2*i+1]*tx_fft1[tx_fft1_px+2*i+1];
    if(t2>r2)r2=t2;
    }
  if(r2 > 0)
    {
    r2=1/sqrt(r2);
    t1=r2;
    if(t1>10)t1=10;
    old_rf1agc[old_display_ptr]=t1;
    rf1agc_pix2=log10(t1)*TX_BAR_SIZE;
    if(r2>10/tx_agc_factor)r2=10/tx_agc_factor;
    if(rf1agc_pix2 < 0)rf1agc_pix2=0;
    }
  else
    {
    rf1agc_pix2=0;
    old_rf1agc[old_display_ptr]=0;
    }
  t1=0;
  for(i=0; i<DISPLAY_HOLD_SIZE; i++)
    {
    if(t1<old_rf1agc[i])t1=old_rf1agc[i];
    }
  if(t1 > 0)t1=20*log10(t1);  
  sprintf(s,"%.1fdB  ",t1);
  lir_text(18+TX_BAR_SIZE/text_width, TX_BAR_RF1AGC_LINE, s);
  tx_bar(16,TX_BAR_RF1AGC_LINE,rf1agc_pix1,rf1agc_pix2);
  rf1agc_pix1=rf1agc_pix2;
  r1=pow(10.0,0.05*tx.rf1_gain)/TX_AGC_THRESHOLD;
  t1=0;
  t2=0;
  prat=0;
  r2*=r1;
  for(i=0; i<tx_fftsize1; i+=2)
    {
    tx_rf1[tx_rf1_pa+i  ]=tx_rf1[tx_rf1_pa+i  ]+r2*tx_fft1[tx_fft1_px+i  ];
    tx_rf1[tx_rf1_pa+i+1]=tx_rf1[tx_rf1_pa+i+1]+r2*tx_fft1[tx_fft1_px+i+1];
    t2=tx_rf1[tx_rf1_pa+i  ]*tx_rf1[tx_rf1_pa+i  ]+
       tx_rf1[tx_rf1_pa+i+1]*tx_rf1[tx_rf1_pa+i+1];
    prat+=t2;
    if(t2>t1)t1=t2;
    if(t2 > 1)
      {
      t2=1/sqrt(t2);
      tx_rf1[tx_rf1_pa+i  ]*=t2;
      tx_rf1[tx_rf1_pa+i+1]*=t2;
      }
    }
  prat/=tx_sizhalf1;
  if(prat>0.00000001)
    {
    prat=10*log10(t1/prat);  
    }
  else
    {
    prat=90;
    }  
  sprintf(s,"[%.1f]  ",prat);
  lir_text(28+TX_BAR_SIZE/text_width, TX_BAR_MICAGC_LINE, s);
  tx_rf1_pa=(tx_rf1_pa+tx_fftsize1)&tx_rfbuf_mask1;
  for(i=0; i<tx_fftsize1; i++)
    {
    tx_rf1[tx_rf1_pa+i]=r2*tx_fft1[tx_fft1_px+tx_fftsize1+i];
    }
  tx_fft1_px=(tx_fft1_px+tx_fft_block1)&tx_rfbuf_mask1;
  old_rf1clip[old_display_ptr]=t1;
  if(t1 > 1000)t1=1000;  
  if(t1 > 1)
    {  
    rf1clip_pix2=log10(t1)*TX_BAR_SIZE/3;
    }
  else
    {
    rf1clip_pix2=0;
    }
  t1=0;
  for(i=0; i<DISPLAY_HOLD_SIZE; i++)
    {
    if(t1<old_rf1clip[i])t1=old_rf1clip[i];
    }
  if(t1 > 1)
    {
    t1=10*log10(t1);  
    }
  else
    {
    t1=0;
    }  
  sprintf(s,"%.1fdB  ",t1);
  lir_text(18+TX_BAR_SIZE/text_width, TX_BAR_RF1CLIP_LINE, s);
  tx_bar(16,TX_BAR_RF1CLIP_LINE,rf1clip_pix1,rf1clip_pix2);
  rf1clip_pix1=rf1clip_pix2;
// ************* Step 5. *************
// At this point we have applied an RF clipper by limiting power in
// Hilbert space. As a consequence we have produced intermodulation
// products outside the desired passband.
// Go to the frequency domain and remove undesired frequencies.
  while( ((tx_rf1_pa-tx_rf1_px+tx_rfbuf_size1)&tx_rfbuf_mask1) > 
                                                           2*tx_fftsize1)
    {                                                  
    t1=0;
    for(i=0; i<tx_sizhalf1; i++)
      {
      t1=tx_rf2[tx_rf1_px+2*i  ]*tx_rf1[tx_rf1_px+2*i  ]+
         tx_rf2[tx_rf1_px+2*i+1]*tx_rf1[tx_rf1_px+2*i+1];
      }
    pb=(tx_rf1_px+2*(tx_fftsize1-1))&tx_rfbuf_mask1;
    j=tx_fftsize1-1;
    for(i=0; i<tx_sizhalf1; i++)
      {
      tx_fft2[tx_fft2_pa+2*i  ]=tx_rf1[tx_rf1_px+2*i  ]*tx_fft_win1[i]; 
      tx_fft2[tx_fft2_pa+2*i+1]=tx_rf1[tx_rf1_px+2*i+1]*tx_fft_win1[i]; 
      tx_fft2[tx_fft2_pa+2*j  ]=tx_rf1[pb  ]*tx_fft_win1[i]; 
      tx_fft2[tx_fft2_pa+2*j+1]=tx_rf1[pb+1]*tx_fft_win1[i]; 
      j--;
      pb-=2;
      }      
    tx_rf1_px=(tx_rf1_px+tx_fftsize1)&tx_rfbuf_mask1;
    fftforward(tx_fftsize1, tx_fftn1, &tx_fft2[tx_fft2_pa], 
                                             tx_fft_table1, tx_fft_permute1);
    lir_sched_yield();
    show_txfft(&tx_fft2[tx_fft2_pa],0,1);
    for(i=0; i<tx.lowest_bin; i++)
      {
      tx_fft2[tx_fft2_pa+2*i  ]=0;
      tx_fft2[tx_fft2_pa+2*i+1]=0;
      }
    for(i=tx.highest_bin+1; i<tx_fftsize1; i++)
      {
      tx_fft2[tx_fft2_pa+2*i  ]=0;
      tx_fft2[tx_fft2_pa+2*i+1]=0;
      }
    tx_fft2_pa=(tx_fft2_pa+tx_fft_block2)&tx_rfbuf_mask2;
    }
// ************* Step 6. *************
// Go back to the time domain and store the signal in tx_rf2.
// Remember that we use sin squared windows and that
// the transforms overlap with 50%.
// Compute power and store the peak power with
// an exponential decay corresponding to a time constant
// of 50 milliseconds.
// We expand the fft size from tx_fftsiz1 to tx_fftsiz2 because
// the fractional resampling that will follow this step needs
// the signal to be oversampled by a factor of four to avoid
// attenuation at high frequencies. The polynomial fitting
// works as a low pass filter and we do not want any attenuation
// at the high side of our passband.
  while(tx_fft2_px != tx_fft2_pa)
    {
  for(i=tx_fftsize1; i<tx_fftsize2; i++)
    {
    tx_fft2[tx_fft2_px+2*i]=0;
    tx_fft2[tx_fft2_px+2*i+1]=0;
    }      
  fftback(tx_fftsize2, tx_fftn2, &tx_fft2[tx_fft2_px], 
                                              tx_fft_table2, tx_fft_permute2);
  lir_sched_yield();
  pmax=0;
  rf2ampl_pa=tx_rf2_pa/2;
  prat=0;
  for(i=0; i<tx_sizhalf2; i++)
    {
//fprintf(dmp,"\nrf2[%d]+=fft2[%d]",tx_rf2_pa+2*i  ,tx_fft2_px+2*i);
//fprintf(dmp,"\nrf2[%d]+=fft2[%d]",tx_rf2_pa+2*i+1,tx_fft2_px+2*i+1);
//fprintf(dmp,"\n%d %f  %f  %f  %f",i,
//      tx_rf2[tx_rf2_pa+2*i  ],tx_fft2[tx_fft2_px+2*i  ],
//      tx_rf2[tx_rf2_pa+2*i+1],tx_fft2[tx_fft2_px+2*i+1]);
      



    tx_rf2[tx_rf2_pa+2*i  ]+=tx_fft2[tx_fft2_px+2*i  ];


    tx_rf2[tx_rf2_pa+2*i+1]+=tx_fft2[tx_fft2_px+2*i+1];
    t2=tx_rf2[tx_rf2_pa+2*i  ]*tx_rf2[tx_rf2_pa+2*i  ]+
       tx_rf2[tx_rf2_pa+2*i+1]*tx_rf2[tx_rf2_pa+2*i+1];
//fprintf(dmp,"\n%d %f",i,sqrt(t2));


    prat+=t2;
    if(t2>pmax)pmax=t2;
    tx_forwardpwr*=txpwr_decay;
    if(tx_forwardpwr<t2)tx_forwardpwr=t2;
    tx_rf2_pwrf[(tx_rf2_pa>>1)+i]=tx_forwardpwr;
    }

// Store an exponential power fall-off with the same time constant, 50 ms,
// in the reverse direction.
  i=tx_sizhalf2-1;
  t1=0;
  while(i >= 0)
    {
    t1*=txpwr_decay;
    if(t1 < tx_rf2_pwrf[(tx_rf2_pa>>1)+i])t1=tx_rf2_pwrf[(tx_rf2_pa>>1)+i];
    tx_rf2_pwrd[(tx_rf2_pa>>1)+i]=t1;
    i--;
    }
  pb=tx_rf2_pa;
  t1=tx_rf2_pwrd[(pb>>1)];
  i=0;
  while( pb != ((int)(tx_rf2_fx+4+tx_rfbuf_size2)&tx_rfbuf_mask2) && i == 0)
    {
    t1*=txpwr_decay;
    pb=(pb+tx_rfbuf_mask2)&tx_rfbuf_mask2;
    if(t1 > tx_rf2_pwrf[(pb>>1)])
      {
      tx_rf2_pwrd[(pb>>1)]=t1;
      }
    else
      {
      i=1;
      }  
    }
  tx_rf2_pa=(tx_rf2_pa+tx_fftsize2)&tx_rfbuf_mask2;
  for(i=0; i<tx_fftsize2; i++)
    {
//fprintf(dmp,"\nrf2[%d]=fft2[%d]",tx_rf2_pa+i,tx_fft2_px+tx_fftsize2+i);

    tx_rf2[tx_rf2_pa+i]=tx_fft2[tx_fft2_px+tx_fftsize2+i];
    }
  tx_fft2_px=(tx_fft2_px+tx_fft_block2)&tx_rfbuf_mask2;
  prat/=tx_sizhalf2;
  if(prat>0.00000001)
    {
    prat=10*log10(pmax/prat);  
    }
  else
    {
    prat=90;
    } 
  old_prat1[old_display_ptr]=prat;
  for(i=0; i<DISPLAY_HOLD_SIZE; i++)
    {
    if(prat>old_prat1[i])prat=old_prat1[i];
    }
  sprintf(s,"[%.1f]  ",prat);
  lir_text(28+TX_BAR_SIZE/text_width, TX_BAR_RF1CLIP_LINE, s);
  if(pmax > 1)
    {  
    pmax=2*log10(pmax);
    if(pmax>1)pmax=1;
    alc_pix2=pmax*TX_BAR_SIZE;
    }
  else
    {
    alc_pix2=0;
    }
  tx_bar(16,TX_BAR_ALC_LINE,alc_pix1,alc_pix2);
  alc_pix1=alc_pix2;
}
// ************* Step 7. *************
// Use the slowly decaying bi-directional peak power as an ALC
// but make sure we allow at least two data blocks un-processed
// to allow the backwards fall-off to become re-calculated.
// Using the slowly varying function for an ALC (= AM modulation)
// will not increase the bandwidth by more than the bandwidth
// of the modulation signal (50 ms or 20 Hz)
  while( ((tx_rf2_pa-tx_rf2_pb+tx_rfbuf_size2)&tx_rfbuf_mask2) > 3*tx_fftsize2)
    {
    pmax=0;
    prat=0;
    for(i=0; i<tx_sizhalf2; i++)
      {
      t1=tx_rf2_pwrd[(tx_rf2_pb>>1)+i];    
      if(t1 > 1)
        {
        t1=1/sqrt(t1);
        tx_rf2[tx_rf2_pb+2*i  ]*=t1;
        tx_rf2[tx_rf2_pb+2*i+1]*=t1;
        }
      t2=tx_rf2[tx_rf2_pb+2*i  ]*tx_rf2[tx_rf2_pb+2*i  ]+
         tx_rf2[tx_rf2_pb+2*i+1]*tx_rf2[tx_rf2_pb+2*i+1];
// We do not want more than 80dB dynamic range.
// any signal below is just rounding errors in muted periods.
      if(t2 < 0.00000001)
        {  
        tx_rf2[tx_rf2_pb+2*i  ]=0;
        tx_rf2[tx_rf2_pb+2*i+1]=0;
        }
      if(t2 > pmax)pmax=t2;
      prat+=t2;
      }
    prat/=tx_sizhalf2;
    old_prat2[old_display_ptr]=prat;
    for(i=0; i<DISPLAY_HOLD_SIZE; i++)
      {
      if(prat>old_prat2[i])prat=old_prat2[i];
      }
    if(prat>0.00000001)
      {
      prat=10*log10(pmax/prat);  
      }
    else
      {
      prat=90;
      }  
    sprintf(s,"[%.1f]  ",prat);
    lir_text(28+TX_BAR_SIZE/text_width, TX_BAR_ALC_LINE, s);
    tx_rf2_pb=(tx_rf2_pb+tx_fftsize2)&tx_rfbuf_mask2;
    }
// ************* Step 8. *************
// In case we have enough signal in the buffer, start the output.
  if(tx_output_flag == 0)
    {
    if( ((tx_rf2_pb-(int)(tx_rf2_fx)+tx_rfbuf_size2)&tx_rfbuf_mask2) > 6*tx_fftsize2)
      {
      tx_output_flag=1;
      tx_if1_pa=0;
      tx_if1_px=0;
      }
    }  
  if(tx_output_flag == 1)
    {
    resample_tx_output();
    }
/*
rf2ampl_pa=0;
pmax=0;
prat=0;
r1=r2=t2=0;
pb=k=j=0;
*/
  test_keyboard();
  if(lir_inkey != 0)
    {
    process_current_lir_inkey();
    if(lir_inkey=='X')goto end_tx_setup;
    if(lir_inkey=='S')
      {
      clear_screen();
      save_tx_parms();
      }
    switch (setup_mode)
      {
      case TX_SETUP_AD:
      if(lir_inkey=='L')
        {
        lir_text(0,TXPAR_INPUT_LINE,"Low:");
        i=lir_get_integer(5,TXPAR_INPUT_LINE,4,1,3000);
        tx.lowest_bin=i/txad_hz_per_bin;
        }
      if(lir_inkey=='H')
        {
        lir_text(0,TXPAR_INPUT_LINE,"High:");
        i=lir_get_integer(6,TXPAR_INPUT_LINE,5,1000,48000);
        tx.highest_bin=i/txad_hz_per_bin;
        }
      if(lir_inkey=='F')
        {
        lir_text(0,TXPAR_INPUT_LINE,"Slope:");
        tx.slope=lir_get_integer(7,TXPAR_INPUT_LINE,4,-50,50);
        }
      if(lir_inkey=='B')
        {
        lir_text(0,TXPAR_INPUT_LINE,"Bass:");
        tx.bass=lir_get_integer(6,TXPAR_INPUT_LINE,4,-50,50);
        }
      if(lir_inkey=='T')
        {
        lir_text(0,TXPAR_INPUT_LINE,"Treble:");
        tx.treble=lir_get_integer(8,TXPAR_INPUT_LINE,4,-50,50);
        }
      break;

      case TX_SETUP_MUTE:
      if(lir_inkey=='F')
        {
        lir_text(0,TXPAR_INPUT_LINE,"Freq domain:");
        tx.mic_f_threshold=lir_get_integer(13,TXPAR_INPUT_LINE,3,0,120);
        }
      if(lir_inkey=='T')
        {
        lir_text(0,TXPAR_INPUT_LINE,"Time domain:");
        tx.mic_t_threshold=lir_get_integer(13,TXPAR_INPUT_LINE,3,0,120);
        }
      break;

      case TX_SETUP_MICAGC:
      if(lir_inkey=='V')
        {
        lir_text(0,TXPAR_INPUT_LINE,"Volume:");
        tx.mic_gain=lir_get_integer(8,TXPAR_INPUT_LINE,3,0,120);
        }
      if(lir_inkey=='T')
        {
        lir_text(0,TXPAR_INPUT_LINE,"Time constant:");
        t1=lir_get_float(15,TXPAR_INPUT_LINE,6,0,10);
        tx.mic_agc_time=100*t1;  
        }
      break;
      
      case TX_SETUP_RF1CLIP:
      if(lir_inkey=='B')
        {
        lir_text(0,TXPAR_INPUT_LINE,"Clipper gain:");
        tx.rf1_gain=lir_get_integer(15,TXPAR_INPUT_LINE,3,-50,120);
        }
      break;
      }
    if(lir_inkey=='+')txtrace_gain*=1.33;
    if(lir_inkey=='-')txtrace_gain/=1.25;

    check_txparms();
    if(lir_inkey=='A')setup_mode=TX_SETUP_MICAGC;
    if(lir_inkey=='M')setup_mode=TX_SETUP_AD;
    if(lir_inkey=='Q')setup_mode=TX_SETUP_MUTE;
    if(lir_inkey=='C')setup_mode=TX_SETUP_RF1CLIP;
    goto restart;
    }
  }
end_tx_setup:;
close_tx_output();
linrad_thread_stop_and_join(THREAD_TX_INPUT);
close_tx_input();
free(txmem_handle);
}

void save_tx_parms(void)
{
int *wgi;
char s[80];
int i, line;
FILE *txpar_file;
wgi=(void*)(&tx.input_speed);
line=4;
txpar_file=fopen("par_tx","w");
if(txpar_file == NULL)
  {
  lirerr(381164);
  return;
  }
tx.check=TXPAR_VERNR;
for(i=0; i<MAX_TXPARM; i++)
  {
  sprintf(s,"%s [%d]",txparm_text[i],wgi[i]);
  fprintf(txpar_file,"%s\n",s);
  lir_text(5,line,s);
  line++;
  }
parfile_end(txpar_file);
sprintf(s,"Parameters saved to par_tx.   %s",press_any_key);
lir_text(10,line+3,s);  
await_keyboard();
}



void tx_input(void)
{
int ia, ib, pa, pb, pc;
int k;
int twosiz;
int nread;
float *z;
twosiz = 2*tx_fftsize1;
thread_status_flag[THREAD_TX_INPUT]=THRFLAG_ACTIVE;
while(!kill_all_flag && thread_command_flag[THREAD_TX_INPUT] == THRFLAG_ACTIVE)
  {
  if(tx.input_bytes == 2)
    {
    nread=read(tx_audio_in,&tx_micbuf_shi[mic_pa],tx_read_bytes); 
    }
  else
    {
    nread=read(tx_audio_in,&tx_micbuf_int[mic_pa],tx_read_bytes); 
    }
  if(nread != tx_read_bytes)
    {
    lirerr(1283);
    goto tx_input_x;
    }  
  mic_pa=(mic_pa+tx_input_block)&tx_micbuf_mask;
  if( ((mic_pa-mic_px+tx_micbuf_size)&tx_micbuf_mask) >= 2*tx_fftsize1)
    {
// Copy the input to tx_fft1 while multiplying with the window function.
    ib=2*tx_fftsize1-1;
    pa=mic_px;
    pb=(pa+ib)&tx_micbuf_mask;
    if(tx.input_bytes == 2)
      {
      if(tx.input_channels==1)
        {
        for( ia=0; ia<tx_fftsize1; ia++)
          {
          tx_tmp[tx_mic_permute[ia]]=(float)(tx_micbuf_shi[pa])*tx_mic_win[ia];
          if(abs(tx_micbuf_shi[pa]) > tx_ad_maxamp)
            {
            tx_ad_maxamp=abs(tx_micbuf_shi[pa]);
            }
          tx_tmp[tx_mic_permute[ib]]=(float)(tx_micbuf_shi[pb])*tx_mic_win[ia];
          if(abs(tx_micbuf_shi[pb]) > tx_ad_maxamp)
            {
            tx_ad_maxamp=abs(tx_micbuf_shi[pb]);
            }
          ib--; 
          pa=(pa+1)&tx_micbuf_mask;
          pb=(pb+tx_micbuf_mask)&tx_micbuf_mask;
          }
        }
      else
        {
// In case there are two channels for the microphone, just use one of them!!
        for( ia=0; ia<tx_fftsize1; ia++)
          {
          tx_tmp[tx_mic_permute[ia]]=tx_micbuf_shi[2*pa]*tx_mic_win[ia];
          if(abs(tx_micbuf_shi[2*pa]) > tx_ad_maxamp)
            {
            tx_ad_maxamp=abs(tx_micbuf_shi[2*pa]);
            }
          tx_tmp[tx_mic_permute[ib]]=tx_micbuf_shi[2*pb]*tx_mic_win[ia];
          if(abs(tx_micbuf_shi[pb]) > tx_ad_maxamp)
            {
            tx_ad_maxamp=abs(tx_micbuf_shi[pb]);
            }
          ib--; 
          pa=(pa+1)&tx_micbuf_mask;
          pb=(pb+tx_micbuf_mask)&tx_micbuf_mask;
          }
        }  
      }
    else
      {
      if(tx.input_channels==1)
        {
        for( ia=0; ia<tx_fftsize1; ia++)
          {
          tx_tmp[tx_mic_permute[ia]]=tx_micbuf_int[pa]*tx_mic_win[ia];
          if(abs(tx_micbuf_int[pa]) > tx_ad_maxamp)
            {
            tx_ad_maxamp=abs(tx_micbuf_int[pa]);
            }
          tx_tmp[tx_mic_permute[ib]]=tx_micbuf_int[pb]*tx_mic_win[ia];
          if(abs(tx_micbuf_int[pb]) > tx_ad_maxamp)
            {
            tx_ad_maxamp=abs(tx_micbuf_int[pb]);
            }
          ib--; 
          pa=(pa+1)&tx_micbuf_mask;
          pb=(pb+tx_micbuf_mask)&tx_micbuf_mask;
          }
        }
      else
        {
// In case there are two channels for the microphone, just use one of them!!
        for( ia=0; ia<tx_fftsize1; ia++)
          {
          tx_tmp[tx_mic_permute[ia]]=tx_micbuf_int[2*pa]*tx_mic_win[ia];
          if(abs(tx_micbuf_int[2*pa]) > tx_ad_maxamp)
            {
            tx_ad_maxamp=abs(tx_micbuf_int[2*pa]);
            }
          tx_tmp[tx_mic_permute[ib]]=tx_micbuf_int[2*pb]*tx_mic_win[ia];
          if(abs(tx_micbuf_int[pb]) > tx_ad_maxamp)
            {
            tx_ad_maxamp=abs(tx_micbuf_int[pb]);
            }
          ib--; 
          pa=(pa+1)&tx_micbuf_mask;
          pb=(pb+tx_micbuf_mask)&tx_micbuf_mask;
          }
        }  
      }
// The microphone signal should be maximum 0x7fff or 0x7fffffffffff
// depending on whether 16 or 32 bit format is selected. 
    mic_px=(mic_px+tx_input_block)&tx_micbuf_mask;
    fft_real_to_hermitian( tx_tmp, 2*tx_fftsize1, tx_fftn1, tx_mic_table);
// Output is {Re(z^[0]),...,Re(z^[n/2),Im(z^[n/2-1]),...,Im(z^[1]).
// Clear those parts of the spectrum that we do not want and
// transfer the rest of the points to tx_fft1  
    k=2*tx.lowest_bin;
    pc=0;
    z=&tx_fft1[tx_fft1_pa];
    while(pc < k)
      {
      z[pc  ]=0;
      z[pc+1]=0;
      pc+=2;
      }
    for(ia=tx.lowest_bin; ia <= tx.highest_bin; ia++)
      {
      z[pc+1]=tx_tmp[ia  ]*mic_filter[ia];
      z[pc  ]=tx_tmp[twosiz-ia]*mic_filter[ia];
      pc+=2;
      }
    while(pc < twosiz)
      {
      z[pc  ]=0;
      z[pc+1]=0;
      pc+=2;
      }
    tx_fft1_pa=(tx_fft1_pa+tx_fft_block1)&tx_rfbuf_mask1;
    lir_sem_post(SEM_TXINPUT);
    }
  }
tx_input_x:;
thread_status_flag[THREAD_TX_INPUT]=THRFLAG_RETURNED;
lir_sem_post(SEM_TXINPUT);
while(thread_command_flag[THREAD_TX_INPUT] != THRFLAG_NOT_ACTIVE)
  {
  lir_sleep(1000);
  }
}

void tx_output(void)
{
thread_status_flag[THREAD_TX_OUTPUT]=THRFLAG_ACTIVE;
while(thread_command_flag[THREAD_TX_OUTPUT] == THRFLAG_ACTIVE)
  {
  }
thread_status_flag[THREAD_TX_OUTPUT]=THRFLAG_RETURNED;
while(thread_command_flag[THREAD_TX_OUTPUT] != THRFLAG_NOT_ACTIVE)
  {
  lir_sleep(1000);
  }
}

void init_txmem(void)
{
int i;
float t1;
// Speech processing is done in both the time domain and in the
// frequency domain.
// The lowest frequency of interest is 150 Hz so it will be
// sufficient to use FFTs that span 10 milliseconds.
// With sin squared windows we use 50% overlap and make a read
// every 5 millisecond. 
// There is a factor of two because the soundcard samples a real-
// valued signal but tx_fftsize1 is for complex transforms.
tx_fftsize1=tx.input_speed*.01;
make_power_of_two(&tx_fftsize1);
tx_fftn1=0;
i=tx_fftsize1;
while(i>1)
  {
  i>>=1;
  tx_fftn1++;
  }
tx_sizhalf1=tx_fftsize1/2;
tx_fft_block_time=(float)(2*tx_fftsize1)/tx.input_speed;
tx_mask=tx_fftsize1-1;
tx_input_block=tx.input_channels*tx_fftsize1;
tx_read_bytes=tx.input_bytes*tx_input_block;
// Allow the input buffer to hold 160 milliseconds of data (or more)
tx_micbuf_size=32*tx_input_block;
tx_micbuf_mask=tx_micbuf_size-1;
init_memalloc(txmem, MAX_TXMEM_ARRAYS);
if(tx.input_bytes == 2)
  {
  mem(1,&tx_micbuf_shi,tx_micbuf_size*sizeof(short int),0);
  }
else
  {
  mem(1,&tx_micbuf_int,tx_micbuf_size*sizeof(int),0);
  }
tx_rfbuf_size1=8*tx_fftsize1;
tx_rfbuf_mask1=tx_rfbuf_size1-1;
tx_resample_ratio=(float)(2*tx.output_speed)/tx.input_speed;
tx_output_upsample=1;
tx_fftsize3=tx_fftsize1;
tx_fftn3=tx_fftn1;
while(tx_resample_ratio > 1.5)
  {
  tx_resample_ratio/=2;
  tx_output_upsample*=2;
  tx_fftsize3*=2;
  tx_fftn3++;
  }
tx_sizhalf3=tx_fftsize3/2;
new_tx_resample_ratio=tx_resample_ratio;
tx_ifbuf_size=2*tx_fftsize3;
mem(2,&tx_fft1,tx_rfbuf_size1*sizeof(float),0);
mem(3,&tx_rf1,tx_rfbuf_size1*sizeof(float),0);
mem(4,&tx_fft2,4*tx_rfbuf_size1*sizeof(float),0);
mem(5,&tx_rf2,4*tx_rfbuf_size1*sizeof(float),0);
// Allow 0.32 seconds in the output buffer (or more).
tx_output_size=tx.output_speed*0.32;
make_power_of_two(&tx_output_size);
tx_output_mask=tx_output_size-1;
if(tx.output_bytes ==2)
  {
  mem(6,&tx_outbuf_shi,tx_output_size*sizeof(short int),0);
  }
else
  {
  mem(6,&tx_outbuf_int,tx_output_size*sizeof(int),0);
  }
mem(7,&tx_tmp,2*tx_fftsize1*sizeof(float),0);
mem(8,&tx_mic_table,tx_fftsize1*sizeof(COSIN_TABLE),0);
mem(9,&tx_mic_permute,2*tx_fftsize1*sizeof(short int),0);
mem(10,&tx_mic_win,(1+tx_fftsize1)*sizeof(float),0);
mem(11,&mic_filter,tx_fftsize1*sizeof(float),0);
mem(12,&tx_fft_table1,tx_fftsize1*sizeof(COSIN_TABLE)/2,0);
mem(13,&tx_fft_permute1,tx_fftsize1*sizeof(short int),0);
mem(14,&tx_fft_win1,(1+tx_sizhalf1)*sizeof(float),0);
mem(15,&tx_rf2_pwrf,4*tx_rfbuf_size1*sizeof(float)/2,0);
mem(16,&tx_rf2_pwrd,4*tx_rfbuf_size1*sizeof(float)/2,0);
mem(17,&tx_if1,4*tx_ifbuf_size*sizeof(float),0);
mem(18,&tx_iftmp1,4*tx_fftsize1*sizeof(float),0);
mem(19,&tx_fft_permute2,4*tx_fftsize1*sizeof(short int),0);
mem(20,&tx_fft_table2,2*tx_fftsize1*sizeof(COSIN_TABLE),0);
mem(21,&tx_fft_win2,(1+4*tx_sizhalf1)*sizeof(float),0);
mem(22,&tx_if2,tx_ifbuf_size*sizeof(float),0);
mem(23,&tx_fft_permute3,tx_fftsize3*sizeof(short int),0);
mem(24,&tx_fft_table3,tx_fftsize3*sizeof(COSIN_TABLE)/2,0);
mem(25,&tx_iftmp2,tx_fftsize3*sizeof(float),0);
if(tx.output_bytes==4)
  {
  mem(26,&tx_out_int,tx.output_channels*tx_fftsize3*sizeof(int)/2,0);
  }
else
  {  
  mem(27,&tx_out_shi,tx.output_channels*tx_fftsize3*sizeof(int)/2,0);
  }
txmem_size=memalloc((int*)(&txmem_handle),"txmem");
if(txmem_size == 0)
  {
  lirerr(1261);
  return;
  }
make_permute(2, tx_fftn1, tx_fftsize1, tx_mic_permute);
make_window(2,tx_fftsize1, 2, tx_mic_win);
make_window(2,tx_sizhalf1, 2, tx_fft_win1);
t1=1/(1.62*tx_fftsize1);
for(i=0; i<tx_sizhalf1; i++)
  {
  tx_fft_win1[i]*=t1;
  }
if(tx.input_bytes != 2)
  {
  for(i=0; i<tx_fftsize1; i++)tx_mic_win[i]/=0x10000;
  }
t1=1/(1.62*0x7fff*tx_sizhalf1);  
for(i=0; i<tx_fftsize1; i++)tx_mic_win[i]*=t1;
make_sincos(2, tx_fftsize1, tx_mic_table); 
init_fft(0,tx_fftn1, tx_fftsize1, tx_fft_table1, tx_fft_permute1);
init_fft(0,tx_fftn3, tx_fftsize3, tx_fft_table3, tx_fft_permute3);
mic_pa=0;
mic_px=0;
tx_fft1_pa=0;
tx_fft1_px=0;
tx_rf1_pa=0;
tx_rf1_px=0;
tx_fft2_pa=0;
tx_fft2_px=0;
tx_rf2_pa=0;
tx_rf2_pb=0;
tx_rf2_fx=0;
tx_output_flag=0;
tx_fft_block1=2*tx_fftsize1;
txad_hz_per_bin=(0.5*tx.input_speed)/tx_fftsize1;
tx_if_phase=0;
tx_if_phstep=0;
check_txparms();
tx_wrblock_no=0;
tx_outbuf_ref=0;
tx_outbuf_current=0;
memcheck(1,txmem,(int*)&txmem_handle);
}
